package com.mangst.sat;

import java.util.List;

public class Expression
{
  /**
   * A List of variables that are part of the entire formula.
   */
  private List<String> variables;

  /**
   * A variable name if this expression represents a single variable, such as
   * "a". Null if the expression represents something like "a & b"
   */
  private String string;

  /**
   * The left-hand expression. For example, the "a" in "a & (b | c)". This will
   * be null if this expression just represents a single variable.
   */
  private Expression left;

  /**
   * The operator of the expression. For example, the "&" in "a & (b | c)". This
   * will be null if this expression just represents a single variable.
   */
  private Operator operator;

  /**
   * The right-hand expression. For example, the "(b | c)" in "a & (b | c)".
   * This will be null if this expression just represents a single variable.
   */
  private Expression right;

  /**
   * True if a NOT should be applied to this expression, false if one shouldn't.
   */
  private boolean bang;

  /**
   * True if this is a sub-expression, false if not--only used for making a
   * nicer toString()
   */
  private boolean subExpression;

  /**
   * The original expression String from the user input.
   */
  private final String origExpression;

  /**
   * Creates a new Expression.
   * @param variables the variables that are in the entire formula, not just
   * this Expression
   * @param expression the formula from which to create the Expression. For
   * example, "a & (b | c)"
   * @throws SATException if there's a problem parsing the formula
   */
  public Expression( List<String> variables, String expression ) throws SATException
  {
    this( variables, expression, false, false );
  }

  /**
   * Creates a new Expression.
   * @param variables the variables that are in the entire formula, not just
   * this Expression
   * @param expression the formula from which to create the Expression. For
   * example, "a & (b | c)"
   * @param bang true if a NOT should be applied to this expression, false if
   * one shouldn't
   * @param subExpression true if this expression is a sub expression, false if
   * not
   * @throws SATException if there's a problem parsing the formula
   */
  private Expression( List<String> variables, String expression, boolean bang, boolean subExpression ) throws SATException
  {
    this.variables = variables;
    origExpression = expression;
    this.bang = bang;
    this.subExpression = subExpression;

    boolean repeat = false;
    do
    {
      int parenDepth = 0;
      int startSubExp = -1, startVar = -1;
      String leftExpression = null, rightExpression = null;
      boolean leftBang = false, rightBang = false;
      boolean leftIsSubExpression = false, rightIsSubExpression = false;
      for ( int i = 0; i < expression.length(); ++i )
      {
        char c = expression.charAt( i );

        //ignore whitespace
        if ( Character.isWhitespace( c ) )
        {
          continue;
        }

        //if not inside a set of parens
        if ( startSubExp < 0 )
        {
          //handle variables
          if ( SATSolver.isVariableChar( c ) )
          {
            if ( startVar < 0 )
            {
              if ( leftExpression == null )
              {
                //start a variable for left-hand expression
                startVar = i;
              }
              else
              {
                //already found the left-hand expression, so make everything else the right-hand expression
                rightExpression = expression.substring( i );
                break;
              }
            }
          }
          else
          {
            if ( startVar >= 0 )
            {
              //end the variable
              int endVar = i;
              if ( leftExpression == null )
              {
                leftExpression = expression.substring( startVar, endVar );
              }
              startVar = -1;
            }
          }

          if ( c == Operator.AND.getSymbol() )
          {
            if ( leftExpression == null )
            {
              throw new SATException( "No left-hand expression: '" + origExpression + "'" );
            }
            operator = Operator.AND;
          }
          else if ( c == Operator.OR.getSymbol() )
          {
            if ( leftExpression == null )
            {
              throw new SATException( "No left-hand expression: '" + origExpression + "'" );
            }
            operator = Operator.OR;
          }
          else if ( c == Operator.NOT.getSymbol() )
          {
            if ( startVar >= 0 )
            {
              throw new SATException( "Can't use " + Operator.NOT.getSymbol() + " in variable names: '" + origExpression + '"' );
            }

            if ( leftExpression == null )
            {
              leftBang = true;
            }
            else
            {
              if ( operator == null )
              {
                throw new SATException( "Invalid use of NOT: '" + origExpression + "'" );
              }
              else
              {
                rightBang = true;
              }
            }
          }
        }

        if ( c == Operator.LPAREN.getSymbol() )
        {
          if ( startSubExp < 0 )
          {
            if ( leftExpression == null )
            {
              //start a sub-expression for left-hand expression
              startSubExp = i + 1;
            }
            else
            {
              //already found the left-hand expression, so make everything else the right-hand expression
              rightExpression = expression.substring( i );
              break;
            }
          }
          parenDepth++;
        }
        else if ( c == Operator.RPAREN.getSymbol() )
        {
          parenDepth--;
          if ( parenDepth == 0 )
          {
            int endSubExp = i;
            leftExpression = expression.substring( startSubExp, endSubExp );
            startSubExp = -1;
            leftIsSubExpression = true;
          }
        }
      } //end for loop

      //if at the end of the string and the variable hasn't been ended, then end it
      if ( startVar >= 0 )
      {
        //end the variable
        int endVar = expression.length();
        if ( leftExpression == null )
        {
          leftExpression = expression.substring( startVar, endVar );
        }
        startVar = -1;
      }

      //check for unmatched parens
      if ( parenDepth > 0 )
      {
        throw new SATException( "Unmatched paraenthesis." );
      }

      repeat = false;
      if ( operator == null )
      {
        //if this.bang is set to true, it should never be set back to false!!
        //the code in this block may run multiple times, so that's why we can't just assign one to the other: bang = leftBang;
        if ( leftBang )
        {
          this.bang = true;
        }

        //then it's a single expression like:
        //"(a | d)" -- note the parens
        //or
        //"a"
        if ( leftIsSubExpression )
        {
          //then remove the parens and do it over again
          expression = leftExpression;
          repeat = true;
        }
        else
        {
          //string = exp.replaceAll( "[\\s!]+", "" ); //remove whitespace and NOTs
          if ( leftExpression != null )
          {
            string = leftExpression.trim();
          }
          else
          {
            throw new SATException( "Problem with '" + origExpression + "'" );
          }
        }
      }
      else
      {
        if ( rightExpression == null )
        {
          throw new SATException( "No right-hand expression: '" + origExpression + "'" );
        }
        left = new Expression( variables, leftExpression, leftBang, leftIsSubExpression );
        right = new Expression( variables, rightExpression, rightBang, rightIsSubExpression );
      }
    }
    while ( repeat );
  }

  /**
   * Evaulates the expression.
   * @param values the value of each variable.
   * @returns true or false, depending on how the expression evaluates
   * @throws SATException if there is a problem evaluating the expression
   */
  public boolean eval( List<Boolean> values ) throws SATException
  {
    boolean b;

    if ( string == null )
    {
      if ( left == null || right == null || operator == null )
      {
        throw new SATException( "Problem at '" + origExpression + "'" );
      }

      if ( operator == Operator.AND )
      {
        b = left.eval( values ) && right.eval( values ); //optimization: Java won't call right.eval() if left.eval() is false
      }
      else if ( operator == Operator.OR )
      {
        b = left.eval( values ) || right.eval( values ); //optimization: Java won't call right.eval() if left.eval() is true
      }
      else
      {
        throw new SATException( "Invalid operator: '" + operator + "'" );
      }
    }
    else
    {
      String variable = string;

      int index = variables.indexOf( variable );
      if ( index < 0 )
      {
        throw new SATException( "Cannot find variable '" + variable + "'" );
      }
      b = values.get( index );
    }

    if ( bang )
    {
      b = !b;
    }

    return b;
  }

  @Override
  public String toString()
  {
    StringBuilder sb = new StringBuilder();

    if ( string == null )
    {
      if ( left.subExpression )
      {
        sb.append( Operator.LPAREN.toString() + left + Operator.RPAREN );
      }
      else
      {
        sb.append( left.toString() );
      }

      sb.append( " " + operator + " " );

      if ( right.subExpression )
      {
        sb.append( Operator.LPAREN.toString() + right + Operator.RPAREN );
      }
      else
      {
        sb.append( right.toString() );
      }
      if ( bang )
      {
        sb.insert( 0, Operator.NOT.toString() + Operator.LPAREN );
        sb.append( Operator.RPAREN );
      }
    }
    else
    {
      if ( bang )
      {
        sb.append( Operator.NOT.toString() );
      }
      sb.append( string );
    }

    return sb.toString();
  }
}
