package com.mangst.sat;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Boolean Satisfiability (SAT) Problem Solver.
 * <p>
 * A SAT formula contains a number of variables whose values can either be true
 * or false. The operators AND, OR, and NOT are used in conjunction with the
 * variables. A SAT solver finds out what the variables have to be in order for
 * the entire formula to evaluate to true.
 * <p>
 * For example, the formula "(a | !b) & c" has three solutions:<br>
 * a=false, b=false, c=true<br>
 * a=true, b=false, c=true<br>
 * a=true, b=true, c=true
 * <p>
 * This is an NP-Complete problem because as the number of inputs grow, the time
 * it takes to solve the problem grows exponentially. For example, with 3
 * variables there are 2^3=8 possible solutions; with 4 variables there are
 * 2^4=16 possible solutions; with 5 there are 2^5=32; and so on. Commercial SAT
 * solvers have clever ways of pruning the search tree in order to more quickly
 * solve the given problem (as opposed to using brute force and trying every
 * single possibility). However, the problem is still technically NP-Complete.
 */
public abstract class SATSolver
{
  public static void main( String args[] ) throws IOException
  {
    final String HELP = "Type a formula and press enter to find its solutions. Type \"quit\" to exit.\nExample formula: (a | !b) & c";
    final String CURSOR = "> ";

    System.out.println( HELP );
    System.out.print( CURSOR );
    BufferedReader in = new BufferedReader( new InputStreamReader( System.in ) );
    String line;
    while ( ( line = in.readLine() ) != null )
    {
      line = line.trim();
      if ( line.length() > 0 )
      {
        if ( line.equalsIgnoreCase( "quit" ) )
        {
          break;
        }

        if ( line.equalsIgnoreCase( "help" ) )
        {
          System.out.println( HELP );
        }
        else
        {
          try
          {
            System.out.println( SATSolver.solve( line ) );
          }
          catch ( Exception e )
          {
            System.out.println( e.getClass().toString().substring( 6 ) + ": " + e.getMessage() );
          }
        }
      }
      System.out.print( CURSOR );
    }
    in.close();
  }

  /**
   * Regular expression used to recognize a variable.
   */
  private static final Pattern variablePattern = Pattern.compile( "[a-zA-Z0-9_]+" );

  /**
   * Finds all solutions to a formula (i.e. all sets of values which result in
   * the formula evaluating to true).
   * @param formula the formula to solve
   * @return a List of all possible solutions
   */
  public static List<Solution> solve( String formula ) throws SATException
  {
    //get all variables in the formula
    List<String> variables = getVariables( formula );

    //compile the expression
    Expression expression = new Expression( variables, formula );

    //get a list of all possible combinations of values that each variable can be (size is equal to 2^n where n = number of variables)
    List<List<Boolean>> combinations = combos( variables.size() );

    //find all solutions
    List<List<Boolean>> solutions = new ArrayList<List<Boolean>>();
    for ( List<Boolean> values : combinations )
    {
      if ( expression.eval( values ) )
      {
        solutions.add( values );
      }
    }

    //create Solution objects to return
    List<Solution> retSolutions = new ArrayList<Solution>( solutions.size() );
    for ( List<Boolean> solution : solutions )
    {
      Map<String, Boolean> values = new TreeMap<String, Boolean>();
      for ( int i = 0; i < solution.size(); ++i )
      {
        values.put( variables.get( i ), solution.get( i ) );
      }
      retSolutions.add( new Solution( values ) );
    }
    return retSolutions;
  }

  /**
   * Gets the name of every variable in the formula.
   * @param formula the formula
   * @return a List of all variable names in the order that they appear in the
   * formula
   */
  public static List<String> getVariables( String formula )
  {
    List<String> variables = new ArrayList<String>();

    Matcher matcher = variablePattern.matcher( formula );
    while ( matcher.find() )
    {
      String s = matcher.group();
      if ( !variables.contains( s ) )
      {
        variables.add( s );
      }
    }

    return variables;
  }

  /**
   * Determines if a character is a valid variable character.
   * @param c the character in question
   * @return true if the character can be used in variables, false if not
   */
  public static boolean isVariableChar( char c )
  {
    return variablePattern.matcher( Character.toString( c ) ).matches();
  }

  /**
   * Recursively finds all possible value sets for a given number of boolean
   * variables. For example, for two variables, there are four possible value
   * sets:<br>
   * [false, false]<br>
   * [false, true]<br>
   * [true, false]<br>
   * [true, true]<br>
   * @param vars the number of variables
   * @throws SATException if vars is less than 1
   * @return a List of all possible value sets
   */
  public static List<List<Boolean>> combos( int vars ) throws SATException
  {
    if ( vars < 1 )
    {
      throw new SATException( "There must be at least one variable." );
    }
    if ( vars == 1 )
    {
      //[[false], [true]]
      List<List<Boolean>> combos = new ArrayList<List<Boolean>>();
      combos.add( new ArrayList<Boolean>()
      {
        {
          add( false );
        }
      } );
      combos.add( new ArrayList<Boolean>()
      {
        {
          add( true );
        }
      } );
      return combos;
    }

    List<List<Boolean>> list = combos( vars - 1 );

    //turn
    //[[false], [true]]
    //into 
    //[[false], [true], [false], [true]]
    int origSize = list.size();
    for ( int i = 0; i < origSize; ++i )
    {
      List<Boolean> l = list.get( i );
      list.add( new ArrayList<Boolean>( l ) );
    }

    //turn
    //[[false], [true], [false], [true]]
    //into
    //[[false, false], [true, false], [false, true], [true, true]]
    for ( int i = 0; i < list.size() / 2; ++i )
    {
      list.get( i ).add( false );
    }
    for ( int i = list.size() / 2; i < list.size(); ++i )
    {
      list.get( i ).add( true );
    }

    return list;
  }
}
